/*
 *  this file is part of sfdroid
 *  Copyright (C) 2015, Franz-Josef Haider <f_haider@gmx.at>
 *  based on harmattandroid by Thomas Perl
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <iostream>
#include <algorithm>
#include <wayland-egl.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <poll.h>

#include "wayland-android-client-protocol.h"
#include "renderer.h"
#include "wayland_helper.h"
#include "sfconnection.h"
#include "utility.h"
#include <QSettings>

using namespace std;

#define QT_SURFACE_EXTENSION_GET_EXTENDED_SURFACE 0

int renderer_t::instances(0);

int renderer_t::init(windowmanager_t &wm)
{
	QSettings  settings("/etc/pure-qml.conf",QSettings::NativeFormat);
    int err = 0;	
	
    frame_callback_ptr = 0;

    GLint configAttribs[] = {
        EGL_SURFACE_TYPE, EGL_WINDOW_BIT,
        EGL_RENDERABLE_TYPE, EGL_OPENGL_ES_BIT,
        EGL_NONE
    };
    EGLConfig egl_cfg;
    EGLint contextParams[] = {EGL_CONTEXT_CLIENT_VERSION, 1, EGL_NONE};
    EGLint numConfigs;


	sleep_frame = settings.value("sleep_frame",0).toInt();
    w_shell_surface = NULL;
    q_extended_surface = NULL;
    xdg_surface = NULL;
    xdg_toplevel = NULL;
    windowmanager = &wm;
    w_surface = NULL;
    egl_surf = NULL;
    w_egl_window = NULL;
    deinit();

#if DEBUG
    cout << "choosing egl config" << endl;
#endif
    if(eglChooseConfig(wayland_helper::egl_display, configAttribs, &egl_cfg, 1, &numConfigs) != EGL_TRUE || numConfigs != 1)
    {
        cerr << "unable to find an EGL Config" << endl;
        err = 7;
        goto quit;
    }

#if DEBUG
    cout << "creating GLES context" << endl;
#endif
    eglBindAPI(EGL_OPENGL_ES_API);
    egl_ctx = eglCreateContext(wayland_helper::egl_display, egl_cfg, NULL, contextParams);
    if(egl_ctx == EGL_NO_CONTEXT)
    {
        cerr << "unable to create GLES context" << endl;
        err = 9;
        goto quit;
    }

#if DEBUG
    cout << "creating surface" << endl;
#endif
    w_surface = wl_compositor_create_surface(wayland_helper::compositor);
    wl_surface_set_buffer_scale(w_surface, wayland_helper::scale);
#if DEBUG
    cout << "creating shell surface" << endl;
#endif
    w_shell_surface = NULL;
    if (wayland_helper::shell)
	w_shell_surface = wl_shell_get_shell_surface(wayland_helper::shell, w_surface);
    else
	cout << "no wl_shell extension" << endl;
    xdg_surface = NULL;
    if (wayland_helper::shell_xdg)
		xdg_surface = xdg_wm_base_get_xdg_surface(wayland_helper::shell_xdg, w_surface);
#if DEBUG
    cout << "getting qt extended surface" << endl;
#endif
    q_extended_surface = NULL;
    if (wayland_helper::q_surface_extension){
	q_extended_surface = (struct qt_extended_surface*)wl_proxy_create((struct wl_proxy *)wayland_helper::q_surface_extension, &wayland_helper::qt_extended_surface_interface);
	wl_proxy_marshal((struct wl_proxy*)wayland_helper::q_surface_extension, QT_SURFACE_EXTENSION_GET_EXTENDED_SURFACE, q_extended_surface, w_surface);
        wl_proxy_add_listener((struct wl_proxy*)q_extended_surface, (void (**)(void))&extended_surface_listener, this);
    }
    else 
	cout << "no qt_surface extension" << endl;
    if(!q_extended_surface)
    {
        //err = 17;
        //goto quit;
    }
    wayland_helper::roundtrip();
    if (xdg_surface){
	xdg_surface_add_listener(xdg_surface, &xdg_surface_listener, this);
	cout << "creating xdg_surface" << endl;
    }
    if (0 && w_shell_surface){
	wl_shell_surface_add_listener(w_shell_surface, &shell_surface_listener, this);
	wl_shell_surface_set_toplevel(w_shell_surface);
    }

    struct wl_region *region;
    region = wl_compositor_create_region(wayland_helper::compositor);
    wl_region_add(region, 0, 0,
                  wayland_helper::width,
                  wayland_helper::height);
    wl_surface_set_opaque_region(w_surface, region);
    wl_region_destroy(region);

#if DEBUG
    cout << "creating wl egl window" << endl;
#endif
    cout << "size " <<   wayland_helper::width << "x"<< wayland_helper::height << endl;
    w_egl_window = wl_egl_window_create(w_surface, wayland_helper::width, wayland_helper::height);
    if(w_egl_window == NULL)
    {
        cerr << "unable to create an egl window" << endl;
        err = 16;
        goto quit;
    }

#if DEBUG
    cout << "creating egl window surface" << endl;
#endif
    egl_surf = eglCreateWindowSurface(wayland_helper::egl_display, egl_cfg, (EGLNativeWindowType)w_egl_window, 0);
    if(egl_surf == EGL_NO_SURFACE)
    {
        cerr << "unable to create an EGLSurface" << endl;
        err = 8;
        goto quit;
    }

#if DEBUG
    cout << "making GLES context current" << endl;
#endif
    if(eglMakeCurrent(wayland_helper::egl_display, egl_surf, egl_surf, egl_ctx) == EGL_FALSE)
    {
        cerr << "unable to make GLES context current" << endl;
        err = 10;
        goto quit;
    }

#if DEBUG
    cout << "setting up gl" << endl;
#endif

#if 1
    glViewport(0, 0, wayland_helper::width, wayland_helper::height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    glOrthof(0, wayland_helper::width, wayland_helper::height, 0, 0, 1);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glEnable(GL_TEXTURE_2D);
    glColor4f(1.f, 1.f, 1.f, 1.f);

    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);

    glGenTextures(1, &dummy_tex);
    glBindTexture(GL_TEXTURE_2D, dummy_tex);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
#endif
#if DEBUG
    // this might even crash if the patch is missing:
    cout << "WARNING: if i crash now the patch is missing, look at " << __FILE__ << "@" << __LINE__ << endl;
    if(glGetString(GL_VERSION) == 0)
    {
        cout << "oh no, gles v1 not working, do you have this patch: https://github.com/mer-hybris/android_frameworks_native/pull/3/files ?" << endl;
        err = 3843;
        goto quit;
    }
#endif

quit:
    return err;
}

void renderer_t::deinit()
{
    for(map<ANativeWindowBuffer*, struct wl_buffer*>::iterator it = buffer_map.begin();it != buffer_map.end();it++)
    {
        wl_buffer_destroy(it->second);
    }
    buffer_map.clear();

    have_focus = false;
    glDeleteTextures(1, &dummy_tex);
    eglMakeCurrent(wayland_helper::egl_display, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT);
    if (egl_surf)
	eglDestroySurface(wayland_helper::egl_display, egl_surf);
    if (w_egl_window)
	wl_egl_window_destroy(w_egl_window);
    if(q_extended_surface)
        wl_proxy_destroy((struct wl_proxy *)q_extended_surface);
    if (w_shell_surface)
	wl_shell_surface_destroy(w_shell_surface);
    if (xdg_toplevel){
	xdg_toplevel_destroy(xdg_toplevel);
	xdg_toplevel = NULL;
    }
    if (xdg_surface){
	xdg_surface_destroy(xdg_surface);
	xdg_surface = NULL;
    }
    if (w_surface)
	wl_surface_destroy(w_surface);
    w_surface = NULL;
    eglDestroyContext(wayland_helper::egl_display, egl_ctx);
}

renderer_t::~renderer_t()
{
    instances--;
}

int renderer_t::draw_raw(void *data, int width, int height, int pixel_format)
{
#if DEBUG
    cout << "draw raw: " << width << " " << height << " " << pixel_format << endl;
#endif
    int err = 0;

    GLuint gl_err = 0;

    float xf = (float)wayland_helper::width / (float)width;
    float yf = 1.f;
    float texcoords[] = {
        0.f, 0.f,
        xf, 0.f,
        0.f, yf,
        xf, yf,
    };

    float vtxcoords[] = {
        0.f, 0.f,
        (float)wayland_helper::width, 0.f,
        0.f, (float)wayland_helper::height,
        (float)wayland_helper::width, (float)wayland_helper::height,
    };

    glVertexPointer(2, GL_FLOAT, 0, &vtxcoords);
    glTexCoordPointer(2, GL_FLOAT, 0, &texcoords);

    glBindTexture(GL_TEXTURE_2D, dummy_tex);

    if(pixel_format == HAL_PIXEL_FORMAT_RGBA_8888 || pixel_format == HAL_PIXEL_FORMAT_RGBX_8888)
    {
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0,
                GL_RGBA, GL_UNSIGNED_BYTE, data);
    }
    else if(pixel_format == HAL_PIXEL_FORMAT_RGB_565)
    {
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width, height, 0,
                GL_RGB, GL_UNSIGNED_SHORT_5_6_5, data);
    }
    else
    {
        cerr << "unhandled pixel format: " << pixel_format << endl;
        err = 3;
        goto quit;
    }
    gl_err = glGetError();
    if(gl_err != GL_NO_ERROR) cerr << "glGetError(): " << gl_err << endl;

    glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

    eglSwapBuffers(wayland_helper::egl_display, egl_surf);

quit:
    return err;
}

int renderer_t::save_screen()
{
    int err = 0;
    int gerr = 0;
    void *buffer_vaddr;

    if(last_screen || !buffer) return 0;
#if DEBUG
    cout << "saving screen" << endl;
#endif
    {
	//void **vaddr;
        uint64_t producer_usage;
        uint64_t consumer_usage;
        gralloc1_rect_t access_region;

        access_region.left = 0;
        access_region.top = 0;
        access_region.width = buffer->width;
        access_region.height = buffer->height;

        android_convertGralloc0To1Usage(GRALLOC_USAGE_HW_COMPOSER/*GRALLOC_USAGE_SW_READ_RARELY*/, &producer_usage, &consumer_usage);
        gerr = gralloc1_lock(gralloc1_device, buffer->handle, producer_usage, consumer_usage, &access_region, &buffer_vaddr, -1);
	/*
        gerr = gralloc_module->lock(gralloc_module, buffer->handle,
        GRALLOC_USAGE_SW_READ_RARELY,
        0, 0, buffer->width, buffer->height,
        &buffer_vaddr);
	*/
    }
    if(gerr)
    {
        cerr << "gralloc lock failed " << gerr << endl;
        err = 3;
        goto quit;
    }

    if(last_screen) free(last_screen);
    last_screen = nullptr;

    if(buffer->format == HAL_PIXEL_FORMAT_RGBA_8888 || buffer->format == HAL_PIXEL_FORMAT_RGBX_8888)
    {
        last_screen = (GLubyte*)malloc(4 * buffer->stride * buffer->height);
        memcpy(last_screen, buffer_vaddr, 4 * buffer->stride * buffer->height);
    }
    else if(buffer->format == HAL_PIXEL_FORMAT_RGB_565)
    {
        last_screen = (GLubyte*)malloc(4 * buffer->stride * buffer->height);
        memcpy(last_screen, buffer_vaddr, 4 * buffer->stride * buffer->height);
    }
    else
    {
        cerr << "unhandled pixel format: " << buffer->format << endl;
        err = 1;
        goto quit;
    }
    {
	int releaseFence = 0;
        err = gralloc1_unlock(gralloc1_device, buffer->handle, &releaseFence);
        close(releaseFence);

	//gralloc_module->unlock(gralloc_module, buffer->handle);
    }

quit:
    return err;
}

int renderer_t::dummy_draw(int stride, int height, int format)
{
#if DEBUG
    cout << "dummy draw" << endl;
#endif

    if(last_screen != nullptr)
    {
        draw_raw(last_screen, stride, height, format);
        free(last_screen);
        last_screen = nullptr;
    }

    return -1;
}

void renderer_t::lost_focus()
{
#if DEBUG
    cout << "losing focus: " << app << " display_id " << display_id << " "<< endl;
#endif
    if (!buffer)
	return;
    if(save_screen() == 0)
    {
        dummy_draw(buffer->stride, buffer->height, buffer->format);
    }
    eglMakeCurrent(wayland_helper::egl_display, EGL_NO_SURFACE, EGL_NO_SURFACE, EGL_NO_CONTEXT);
#if 0
    if (multiwindow && xdg_toplevel)
		xdg_toplevel_set_minimized(xdg_toplevel);
#endif
    have_focus = false;
}

void renderer_t::gained_focus(int display_id)
{
	cout << "gained focus " << app << endl;
    if (0 &&w_shell_surface)
		wl_shell_surface_set_toplevel(w_shell_surface);

    if (!xdg_toplevel){
		xdg_toplevel = xdg_surface_get_toplevel(xdg_surface);
		xdg_toplevel_add_listener(xdg_toplevel,
									  &xdg_toplevel_listener, this);
    }
    if (display_id == 0){
		xdg_toplevel_set_app_id(xdg_toplevel, "android");
		if (!multiwindow)
			xdg_toplevel_set_title(xdg_toplevel, "android");
		else {
			xdg_toplevel_set_title(xdg_toplevel, app.c_str());
		}
    }
    else {
		xdg_toplevel_set_app_id(xdg_toplevel, "tpm_android");
		if (!multiwindow)
			xdg_toplevel_set_title(xdg_toplevel, "tpm_android");
		else {
			xdg_toplevel_set_title(xdg_toplevel, ("tpm_"+app +"##"+ std::to_string(stack_id_sf)).c_str());
		}
    }
    wl_surface_commit(w_surface);
    eglMakeCurrent(wayland_helper::egl_display, egl_surf, egl_surf, egl_ctx);
    have_focus = true;
    cout << "have focus: " <<app << endl;
}

int renderer_t::recreate()
{
    int ret;

    cout << "recreate" << endl;
    deinit();
    ret = init(*windowmanager);
    have_focus = true;
    return ret;
}

bool renderer_t::is_active()
{
    return have_focus;
}

int renderer_t::render_buffer(ANativeWindowBuffer *the_buffer, buffer_info_t &info,bool isdummy)
{
    bool slowmode = false;
#if DEBUG
    cout << "rendering buffer in: " << app << " display_id " << display_id << endl;
#endif
    buffer = the_buffer;

    //info.width = 350; info.height = 600 ;
    if(buffer_map.find(the_buffer) == buffer_map.end())
    {
        struct wl_buffer *w_buffer;
        struct wl_array ints;
        int *the_ints;
        struct android_wlegl_handle *wlegl_handle;

	//cout << "rendering buffer in: numInts " << app << " display_id " << display_id << " " << buffer->handle->numInts << " " << buffer->handle->numFds << " " << info.pixel_format << endl;
        wl_array_init(&ints);
        the_ints = (int*)wl_array_add(&ints, buffer->handle->numInts * sizeof(int));
        memcpy(the_ints, buffer->handle->data + buffer->handle->numFds, buffer->handle->numInts * sizeof(int));
        wlegl_handle = android_wlegl_create_handle(wayland_helper::a_android_wlegl, buffer->handle->numFds, &ints);
        wl_array_release(&ints);

        for (int i = 0; i < buffer->handle->numFds; i++)
        {
            android_wlegl_handle_add_fd(wlegl_handle, buffer->handle->data[i]);
        }
	//cout << "rendering buffer in: 3" << app <<" "<< info.stride << endl;
        w_buffer = android_wlegl_create_buffer(wayland_helper::a_android_wlegl,info.width, info.height, info.stride, info.pixel_format, GRALLOC_USAGE_HW_RENDER , wlegl_handle);
        android_wlegl_handle_destroy(wlegl_handle);

        wl_buffer_add_listener(w_buffer, &w_buffer_listener, this);

        buffer_map[the_buffer] = w_buffer;
    }
	//here was
	
    
    //cout << "bpz d " << display_id << endl;
    if(!istop/*!have_focus*/)
    {
        // lost focus due to keyboard leave
		//cout << "not top skipping " << display_id << endl;
		//usleep(4000);
        //return 0;
    }
    //cout << "bp d " << display_id << " counter "<< ++counter << endl;
    if (1/*slowmode*/){//we need this in case at least of slowmode(sleep) even wl_surface_frame is enough(client->waylserv)
		frame_callback_ptr = wl_surface_frame(w_surface);
		wl_callback_add_listener(frame_callback_ptr, &w_frame_listener, this);
    }

    
    wl_surface_attach(w_surface, buffer_map[the_buffer], 0, 0);
    wl_surface_damage(w_surface, 0, 0, info.width, info.height);
    wl_surface_commit(w_surface);
	if (0)
	{
		int ret = 0;
		uint64_t t1,t2;

		t1 =  get_current_time_mcs();
		wl_display_roundtrip(wayland_helper::display);
		while(frame_callback_ptr && ret != -1){
			ret = wl_display_dispatch(wayland_helper::display);
		}
		t2 =  get_current_time_mcs();
		fprintf(stderr,"p %lld/%lld  disp %d\n",(t2 - t1)/1000,(t1-lt)/1000,display_id);
		lt = t1;
	}
	if (1 && arender[display_id] && (arender[display_id]->app.find("videolan") == string::npos)){//hmm build-in player do not need it
		uint64_t t1,t2;
		int ret1;

        //ret = wl_display_dispatch(wayland_helper::display);
		
		t1 =  get_current_time_mcs();
		ret1 = wl_display_roundtrip(wayland_helper::display);
		t2 =  get_current_time_mcs();
		//fprintf(stderr,"p %lld/%lld  disp %d\n",(t2 - t1)/1000,(t1-lt)/1000,display_id);
		lt = t1;
		//cout << "ap "<< ret1 << endl;
    }
    else
		slowmode = true;
	
    //cout << "rendering buffer in: 2 " << app <<" "<<info.width<< " " << info.height <<  buffer_map[the_buffer] << endl;
#if DEBUG
    cout << "rendering buffer in2: " << app << " display_id " << display_id << endl;
#endif
    if(last_screen)
		free(last_screen);
    last_screen = nullptr;
    {
#if 0
        struct pollfd fdp;
	int result;


	fdp.fd = event_fd;
	fdp.events = POLLIN;

	ResetEvent(event_fd);
	cout << "bp" << endl;
	result = poll(&fdp,1,16);
	cout << "ap" << endl;
	if (fdp.revents & POLLIN)
	    ResetEvent(event_fd);
#endif
    if (slowmode && !isdummy){
		if (display_id == 0){
			usleep(sleep_frame/*16000*/);//dispatch can take up to 90 msec on screen 1 and then 0 screen 2 , so let us sleep 1 frame
		}
		else {
		  //if(istop)
		  //usleep(8000);//hmm if not in clone mode we should  sleep
		}
    }
#if 0
	wl_display_flush(wayland_helper::display);
	isbreak = false;
	while(1){
	    //cout << "1 d " << display_id << endl;
	    //wl_display_dispatch(wayland_helper::display);
	    //cout << "2 d " << display_id << endl;
	    //usleep(4000);
	    if (isbreak)
		break;
	}
#endif
    }
    return 0;
}

void renderer_t::buffer_release(void *data, struct wl_buffer *buffer)
{
#if DEBUG
    cout << "buffer release" << endl;
#endif

    // we're cleaning in deinit()
}

void renderer_t::shell_surface_ping(void *data, struct wl_shell_surface *shell_surface, uint32_t serial)
{
#if DEBUG
    cout << "shell surface ping " << endl;
#endif
    wl_shell_surface_pong(shell_surface, serial);
}

void renderer_t::shell_surface_configure(void *data, struct wl_shell_surface *shell_surface, uint32_t edges, int32_t width, int32_t height)
{
#if DEBUG
    cout << "shell surface configure " << width << " " << height <<endl;
#endif
    renderer_t *self = (renderer_t*)data;
    wl_egl_window_resize(self->w_egl_window, width, height, 0, 0);
}

void renderer_t::shell_surface_popup_done(void *data, struct wl_shell_surface *shell_surface)
{
#if DEBUG
    cout << "shell surface popup done" << endl;
#endif
}

void renderer_t::handle_onscreen_visibility(void *data, struct qt_extended_surface *qt_extended_surface, int32_t visible)
{
#if DEBUG
    renderer_t *renderer = (renderer_t*)data;
    cout << "qt_extended_surface handle onscreen visibility @ "<< renderer->app << endl;
#endif
    // handled in keyboard leave/enter
}

void renderer_t::handle_set_generic_property(void *data, struct qt_extended_surface *qt_extended_surface, const char *name, struct wl_array *value)
{
#if DEBUG
    cout << "qt_extended_surface handle set generic property" << endl;
#endif
}

void renderer_t::handle_close(void *data, struct qt_extended_surface *qt_extended_surface)
{
#if DEBUG
    cout << "qt_extended_surface handle close" << endl;
#endif

    renderer_t *renderer = (renderer_t*)data;
    renderer->windowmanager->handle_close(renderer->w_surface);
}
//xdg
void renderer_t::handle_surface_configure(void *data, struct xdg_surface *surface,uint32_t serial)
{
	//struct window *window = data;

	xdg_surface_ack_configure(surface, serial);
	cout << "xdg shell surface configure " << serial << endl;
}
bool renderer_t::is_ignore_config(int idi)
{
    auto it = std::find_if(confids.begin(),confids.end(),[&](int id)
    	  {
	    return (id == idi);
	  });
    if (it != confids.end()){
	cout << "ignoring configure id " << idi << endl;
	return true;
    }
    return false;
}
void renderer_t::handle_toplevel_configure(void *data, struct xdg_toplevel *toplevel,
			  int32_t width, int32_t height,
			  struct wl_array *states)
{
	renderer_t *renderer = (renderer_t*)data;
	auto *xdgStates = static_cast<uint32_t *>(states->data);
	size_t numStates = states->size / sizeof(uint32_t); 

	cout << "xdg toplevel surface configure " << width << " " << height << " display_id: "<< renderer->display_id << endl;
	if (width && height){
	    wl_egl_window_resize(renderer->w_egl_window, width, height, 0, 0);
	}
	for (size_t i = 0; i < numStates; i++) 
	    cout << "xdg toplevel surface configure states " << (int)xdgStates[i] << endl;
	if (numStates == 3 && ((int)xdgStates[0] == -3 && ((int)xdgStates[1] == -4 || (int)xdgStates[1] == -5 || (int)xdgStates[1] == -7))){//got focus
	    if (renderer->is_ignore_config((int)xdgStates[2]))
			return;
	    renderer->istop = true;
	    renderer->have_focus = true;
	    block_input(renderer->display_id,false);
	    if (renderer->multiwindow){
			parender[renderer->display_id] = arender[renderer->display_id];
			arender[renderer->display_id] = renderer;
			cout << "set arender to "<< renderer->app  << " d " << renderer->display_id;
			if ((int)xdgStates[1] == -4 || (int)xdgStates[1] == -7){//-7 previos was native window
				to_front(renderer->app.c_str());
				start_running_process();
				if ((int)xdgStates[1] == -7)
					arender[renderer->display_id]->isprevnative = true;
				else
					arender[renderer->display_id]->isprevnative = false;
			}
	    }
	    renderer->confids.push_back((int)xdgStates[2]);
	}
	if (numStates == 3 && ((int)xdgStates[0] == -1 && (int)xdgStates[1] == -2)){//lost focus
	    if (renderer->is_ignore_config((int)xdgStates[2]))
			return;
	    renderer->istop = false;
	    renderer->have_focus = false;
	    block_input(renderer->display_id,true);
	    renderer->confids.push_back((int)xdgStates[2]);
	}
	if (numStates == 4 && (int)xdgStates[0] == -6){//resize
	    if (renderer->is_ignore_config((int)xdgStates[3]))
		return;
	    int32_t h = (int32_t)xdgStates[1];
	    int32_t istopa = (int32_t)xdgStates[2];
	    resize_window(renderer->display_id,h);
	    if (istopa)
	        block_input(renderer->display_id,false);
	    renderer->confids.push_back((int)xdgStates[3]);
	}
}
void renderer_t::handle_toplevel_close(void *data, struct xdg_toplevel *xdg_toplevel)
{
	renderer_t *renderer = (renderer_t*)data;

	cout << "xdg toplevel close" << endl;
	renderer->windowmanager->handle_close(renderer->w_surface);

}
void renderer_t::frame_callback(void *data, struct wl_callback *callback, uint32_t time)
{
    renderer_t *renderer = (renderer_t*)data;

    //renderer->isbreak = true;
    //cout << "ap d " << renderer->display_id << " counter "<< renderer->counter << endl;
    renderer->frame_callback_ptr = 0;
    wl_callback_destroy(callback);
}

